/*
 * $Id: schedule.c,v 1.6 1998/07/17 18:35:00 masaki Exp $
 */		

#include <mrt.h>
#include <string.h>
#include <stdarg.h>
#include <schedule.h>
#include <timer.h>

schedule_master_t *SCHEDULE_MASTER;

void Delete_Event (event_t *event) {Delete ( (void *)event);}

/* init_schedule
 */
int init_schedules (trace_t *tr) {
  assert (SCHEDULE_MASTER == NULL);
  SCHEDULE_MASTER = New (schedule_master_t);
  SCHEDULE_MASTER->ll_schedules = LL_Create (0);
  SCHEDULE_MASTER->trace = tr;
  pthread_mutex_init (&SCHEDULE_MASTER->mutex_lock, NULL);
  return (1);
}


/* New_Schedule
 */
schedule_t *New_Schedule (char *description, trace_t *tr) {
   schedule_t *tmp;

   tmp = New (schedule_t);
   tmp->ll_events = LL_Create (LL_DestroyFunction, Delete_Event, 0);

   if (description == NULL)
      description = "unknown";
   tmp->description = strdup (description);

   /* this trace condition is delegated from the caller */
   tmp->trace = tr;

   /* tmp->self = pthread_self (); */
   pthread_cond_init (&tmp->cond_new_event, NULL);
   pthread_mutex_init (&tmp->mutex_lock, NULL);
   pthread_mutex_init (&tmp->mutex_cond_lock, NULL);

   pthread_mutex_lock (&SCHEDULE_MASTER->mutex_lock);
   LL_Add (SCHEDULE_MASTER->ll_schedules, tmp);
   pthread_mutex_unlock (&SCHEDULE_MASTER->mutex_lock);

   return (tmp);

}


int schedule_event (schedule_t *schedule, void (*call_fn)(), 
		    int narg, ...) {
   va_list ap;
   void *arg1, *arg2, *arg3, *arg4;

   va_start (ap, narg);

   if (narg <= 0) {
         return (schedule_event2 ("No Description", schedule, call_fn, 0, NULL));
   }
   arg1 = va_arg (ap, void *);
   if (narg == 1) {
     return (schedule_event2 ("No Description", schedule, call_fn, 1, arg1));
   }
   arg2 = va_arg (ap, void *);
   if (narg == 2) {
         return (schedule_event2 ("No Description", schedule, call_fn, 2, arg1, arg2));
   }
   arg3 = va_arg (ap, void *);
   if (narg == 3) {
         return (schedule_event2 ("No Description", schedule, call_fn, 
				  3, arg1, arg2, arg3));
   }
   arg4 = va_arg (ap, void *);
   if (narg == 4) {
         return (schedule_event2 ("No Description", schedule, call_fn, 
				  4, arg1, arg2, arg3, arg4));
   }
   assert (0);
   return (0);
}


/* schedule_event
 */
int schedule_event2 (char *description, schedule_t *schedule, 
		     void (*call_fn)(), int narg, ...) {

   va_list    	ap;
   event_t	*event;

   event = New (event_t);

   /* I expect that description is in a static strage */
   event->description = description;

   va_start (ap, narg);

   if (narg >= 1)
      event->arg1 = va_arg (ap, void *);
   if (narg >= 2)
      event->arg2 = va_arg (ap, void *);
   if (narg >= 3)
      event->arg3 = va_arg (ap, void *);
   if (narg >= 4)
      event->arg4 = va_arg (ap, void *);
   if (narg >= 5)
      assert (0);

   va_end (ap);
   
   event->call_fn = call_fn;
   event->narg = narg;

   /* obtain lock */
   pthread_mutex_lock (&schedule->mutex_cond_lock);

   LL_Add (schedule->ll_events, event);
   schedule->new_event_flag ++;

   pthread_cond_signal (&schedule->cond_new_event);

   if (event->description)
      trace (TR_THREAD, schedule->trace, 
	  "THREAD Event %s scheduled (%d events) for %s\n", 
	  event->description, LL_GetCount (schedule->ll_events),
	  schedule->description);
   else
      trace (TR_THREAD, schedule->trace, 
	  "THREAD Event %x scheduled (%d events) for %s\n", 
	  call_fn, LL_GetCount (schedule->ll_events),
	  schedule->description);

   /* release lock */
   pthread_mutex_unlock (&schedule->mutex_cond_lock);

   if (LL_GetCount (schedule->ll_events) > 5000) {
      trace (TR_THREAD, schedule->trace, "THREAD Over 5000 events !!!!!!");
      trace (TR_THREAD, schedule->trace, "THREAD *** SCHEDULE OVERFLOW ***");
      return (-1);
   }
   
   return (1);
}


int
schedule_event_dispatch (event_t *event)
{
   int narg = event->narg;
   void (*call_fn)() = event->call_fn;
   void *arg1 = event->arg1;
   void *arg2 = event->arg2;
   void *arg3 = event->arg3;
   void *arg4 = event->arg4;

   Delete_Event (event); /* in case function doesn't return */

   if (narg == 0)
      call_fn ();
   else if (narg == 1)
      call_fn (arg1);
   else if (narg == 2)
      call_fn (arg1, arg2);
   else if (narg == 3)
      call_fn (arg1, arg2, arg3);
   else if (narg == 4)
      call_fn (arg1, arg2, arg3, arg4);
   else
      assert (0);

   return (1);
}


/* schedule_wait_for_event
 */
int schedule_wait_for_event (schedule_t *schedule) {
   event_t *event;

#ifndef HAVE_LIBPTHREAD
	return (1);
#endif

   pthread_mutex_lock (&schedule->mutex_cond_lock);

   while (schedule->new_event_flag == 0)
      pthread_cond_wait (&schedule->cond_new_event, 
			 &schedule->mutex_cond_lock);

   /*trace (TR_THREAD, schedule->trace, "Processing event queue for %s", 
	  schedule->description);*/

   event = (event_t *) LL_GetHead (schedule->ll_events);

   if (event != NULL) {
     LL_RemoveFn (schedule->ll_events, event, NULL);
     schedule->new_event_flag --;
   }
   else {
     trace (TR_THREAD, schedule->trace, 
	    "THREAD NULL event for %s -- strange...", 
	    schedule->description);
   }

   /* unlock -- we removed the event from the schedule quere */
   pthread_mutex_unlock (&schedule->mutex_cond_lock);

   if (event == NULL) {return (-1);}

   schedule_event_dispatch (event);
   return (1);
}


/* intervene_thread
 * Halt a thread so we can gain access to its private data space.
 * This is a convenience function mainly used if config and show output
 * routines. Easier for the uii or config thread to retain control than 
 * have to negotiate output, etc.
 *
 */
void intervene_thread (pthread_mutex_t *mutex, pthread_mutex_t *mutex_cond_lock, 
		       pthread_cond_t *cond) {

  /* signal cond lock */
  pthread_mutex_lock (mutex_cond_lock);
  pthread_cond_signal (cond);

  /* wait on mutex  -- other process is running now */
  pthread_mutex_lock (mutex);
  
}

pthread_mutex_t *intervene_thread_start (schedule_t *schedule) { 
  pthread_mutex_t *mutex, *mutex_cond_lock;
  pthread_cond_t *cond;
  
  mutex = New (pthread_mutex_t);
  mutex_cond_lock = New (pthread_mutex_t);
  cond = New (pthread_cond_t);


  pthread_mutex_init (mutex, NULL);
  pthread_mutex_init (mutex_cond_lock, NULL);
  pthread_cond_init (cond, NULL);

  /* grab mutex */
  pthread_mutex_lock (mutex);

  pthread_mutex_lock (mutex_cond_lock);
  schedule_event2 ("Intervene", schedule, intervene_thread, 3, 
		   mutex, mutex_cond_lock, cond);
  pthread_cond_wait (cond, mutex_cond_lock);

  /* cleanup */
  pthread_mutex_destroy (mutex_cond_lock);
  
  /* return and do our thing */
  return (mutex);
}

void intevene_thread_end (pthread_mutex_t *mutex) {
  /* unlock mutex */
  pthread_mutex_unlock (mutex);

  pthread_mutex_destroy (mutex);
}







/* clear_schedule
 */
int clear_schedule (schedule_t *schedule) {

#ifdef HAVE_LIBPTHREAD
   /* obtain lock */
   if (pthread_mutex_trylock (&schedule->mutex_cond_lock) != 0) {
      trace (TR_THREAD, schedule->trace, 
	     "THREAD Going to block while clearing schedule for %s\n",
	     schedule->description);
      pthread_mutex_lock  (&schedule->mutex_cond_lock);
   }
#endif /* HAVE_LIBPTHREAD */

   LL_Clear (schedule->ll_events);
   schedule->new_event_flag = 0;
   trace (TR_THREAD|NORM, schedule->trace, 
	  "THREAD Cleared schedule for %s\n", schedule->description);

#ifdef HAVE_LIBPTHREAD
   pthread_mutex_unlock  (&schedule->mutex_cond_lock);
#endif /* HAVE_LIBPTHREAD */
   return (1);
}


/* delete_schedule
 */
int delete_schedule (schedule_t *schedule) {

   pthread_mutex_lock (&SCHEDULE_MASTER->mutex_lock);
   LL_RemoveFn (SCHEDULE_MASTER->ll_schedules, schedule, 0);
   pthread_mutex_unlock (&SCHEDULE_MASTER->mutex_lock);
   pthread_mutex_lock  (&schedule->mutex_cond_lock);
   LL_Destroy (schedule->ll_events);
   trace (TR_THREAD|NORM, schedule->trace, 
	  "THREAD Deleted schedule for %s\n", schedule->description);
   Delete (schedule->description);
   pthread_cond_destroy (&schedule->cond_new_event);
   pthread_mutex_destroy (&schedule->mutex_lock);
   pthread_mutex_destroy (&schedule->mutex_cond_lock);
   Delete (schedule);
   return (1);
}


#ifndef HAVE_LIBPTHREAD

/* process_all_schedules
 */
int process_all_schedules () {
  schedule_t *schedule;
  event_t *event = NULL;

  LL_Iterate (SCHEDULE_MASTER->ll_schedules, schedule) {
    if (schedule->new_event_flag > 0) {
      event = (event_t *) LL_GetHead (schedule->ll_events);
      break;
    }
  }

  if (event == NULL) {
    return (0);
  }

  LL_RemoveFn (schedule->ll_events, event, NULL);
  schedule->new_event_flag --;
  if (schedule->self)
    set_thread_id (schedule->self);
  if (event->description)
    trace (TR_THREAD, schedule->trace, 
      "THREAD Event %s now run (%d events left) for %s\n", 
      event->description, LL_GetCount (schedule->ll_events),
      schedule->description);
  else
    trace (TR_THREAD, schedule->trace, 
      "THREAD Event %x now run (%d events left) for %s\n", 
      event->call_fn, LL_GetCount (schedule->ll_events),
      schedule->description);

  schedule_event_dispatch (event);
  return (1);
}


static int current_thread_id = 1;

pthread_t get_thread_id () {
   return (current_thread_id);
}

pthread_t set_thread_id (pthread_t id) {
   current_thread_id = id;
   return (current_thread_id);
}
#endif /* HAVE_LIBPTHREAD */


int schedule_count (schedule_t *schedule) {
  int num;

  if (schedule == NULL) {return (-1);}
  /* obtain lock */
  pthread_mutex_lock (&schedule->mutex_cond_lock);
  num = LL_GetCount (schedule->ll_events);
  pthread_mutex_unlock (&schedule->mutex_cond_lock);

  return (num);
}
